#include <pspkernel.h>
#include <pspdisplay.h>
#include <pspdebug.h>
#include <pspgu.h>
#include <pspgum.h>
#include <pspctrl.h>
#include <string.h>

PSP_MODULE_INFO("bossbullet3", PSP_MODULE_USER, 1, 1);
static unsigned int __attribute__((aligned(16))) list[262144];
extern unsigned char spr_start[]; // テクスチャ参照
#define BUF_WIDTH (512)
#define SCR_WIDTH (480)
#define SCR_HEIGHT (272)
#define CHR_COUNT 256 // スプライト最大数256を指定

// 描画用Vertex構造体
typedef struct Vertex{
	unsigned short u, v;
	unsigned short color;
	short x, y, z;
}Vertex;
struct Vertex* vertices;

// テクスチャ割り当てをプリプロセッサで定義
#define MY_SHIP 0
#define MY_BOSS 6
#define MY_EFFECT 10
#define MY_BULLET 20
#define MY_ENEMY_BULLET 50

// 敵弾の最大描画数(とりあえず100発まで)
#define MY_ENEMY_BULLET_MAX 100

// グローバル変数定義
unsigned long i=0,j=0;	// 汎用
SceCtrlData pad;		// パッド情報格納用
unsigned long m_ulFrameCount=0; // フレームカウント変数

// 自機＆バレット用パラメータ定義
unsigned int m_iShip_x,m_iShip_y;
static int m_iBullet_x[11],m_iBullet_y[11],m_iBullet_flag[11];
static int m_iBullet_spd;
// ボス用パラメータ定義。全4ステージを想定してボス4体分
static int m_iStage=0;						// ひとまずステージ1を選択した場合を想定
static int m_iBoss_x[5]={360,360,360,360};	// ボスX座標
static int m_iBoss_y[5]={104,104,104,104};	// ボスY座標
static int m_iBoss_spd[5]={2,2,2,2};		// ボス移動スピード
static int m_iBoss_hp[5]={100,100,100,100};	// ボスの耐久性
static int m_iBoss_u[5]={  0,  0,  0,  0};	// ボスのスプライト位置指定用u
static int m_iBoss_v[5]={ 32,  0,  0,  0};	// ボスのスプライト位置指定用v
static int m_iBoss_MoveX=0,m_iBoss_MoveY=0;	// 移動関連パラメータ
// エフェクト用パラメータ定義
static int m_iEffect_x[6],m_iEffect_y[6];	// エフェクト用X座標
static int m_iEffect_time[6];				// エフェクトアニメーション用
// 敵弾用パラメータ定義。
double m_dEnemyBullt_x[MY_ENEMY_BULLET_MAX+1];				// 敵弾X座標
double m_dEnemyBullt_y[MY_ENEMY_BULLET_MAX+1];				// 敵弾Y座標
static int m_iEnemyBullet_Type[MY_ENEMY_BULLET_MAX+1];		// 敵弾タイプ
double m_dENemyBullet_Spd[MY_ENEMY_BULLET_MAX+1];			// 敵弾スピード
static int m_iEnemyBullet_Num[MY_ENEMY_BULLET_MAX+1];		// 敵弾画像番号
static int m_iEnemyBullet_Hitbox[MY_ENEMY_BULLET_MAX+1];	// 敵弾当たり判定有無(0=スルー)
static int m_iEnemyBullet_Power[MY_ENEMY_BULLET_MAX+1];		// 敵弾耐久値
static int m_iEnemyBullet_Autodec[MY_ENEMY_BULLET_MAX+1];	// 敵弾自動減衰値
// その他グローバル変数
static int m_iGameoverFlag=0;				// ゲームオーバーフラグ




// 変数初期化自作関数
void MyInitParam(){
	m_iShip_x=32; m_iShip_y=120; // 自機の初期座標を左中央付近へ
	memset(m_iBullet_x,-64,sizeof(m_iBullet_x)); // バレットの位置を(-64,-64)に指定
	memset(m_iBullet_y,-64,sizeof(m_iBullet_y));
	memset(m_iBullet_flag,0,sizeof(m_iBullet_flag)); // 射出フラグは0リセット
	memset(m_iEffect_time,-1,sizeof(m_iEffect_time)); // エフェクトフラグもゼロクリア
	memset(m_dEnemyBullt_x,0,sizeof(m_dEnemyBullt_x)); // 敵弾全座標リセット
	memset(m_dEnemyBullt_y,0,sizeof(m_dEnemyBullt_y)); // 敵弾全座標リセット
	memset(m_iEnemyBullet_Type,-1,sizeof(m_iEnemyBullet_Type)); // 敵弾タイプを全てリセット
	m_iBullet_spd = 8; // 自機バレットの初期速度は8にセット
	m_ulFrameCount = 0; // フレームカウントをゼロに。
	sceCtrlSetSamplingCycle(0);	// パッド情報も初期化
	sceCtrlSetSamplingMode(PSP_CTRL_MODE_ANALOG);
}

// gu初期化自作関数
void MyInitGu(){
	sceGuInit(); sceGuStart(GU_DIRECT, list);
	sceGuDrawBuffer(GU_PSM_4444, (void*)0, BUF_WIDTH);
	sceGuDispBuffer(SCR_WIDTH, SCR_HEIGHT, (void*)0x44000, BUF_WIDTH);
	sceGuOffset(2048 - (SCR_WIDTH/2), 2048 - (SCR_HEIGHT/2));
	sceGuViewport(2048, 2048, SCR_WIDTH, SCR_HEIGHT);
	sceGuScissor(0, 0, SCR_WIDTH, SCR_HEIGHT);
	sceGuEnable(GU_SCISSOR_TEST); sceGuEnable(GU_TEXTURE_2D);


// テクスチャデータのセット
	sceGuTexMode(GU_PSM_4444, 0, 0, GU_FALSE);
	sceGuTexImage(0, 256, 256, 256, spr_start);
	sceGuTexFunc(GU_TFX_REPLACE, GU_TCC_RGBA);
	sceGuTexFilter(GU_NEAREST, GU_NEAREST);
	// 透過ルーチン
	sceGuEnable(GU_BLEND);
	sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0); 

	
	// 設定完了
	sceGuFinish(); sceGuSync(0, GU_SYNC_FINISH);
	sceDisplayWaitVblankStart(); sceGuDisplay(GU_TRUE);
	// ディスプレイリストからメモリ割り当て(CHR_COUNT利用)
	vertices = (struct Vertex*)sceGuGetMemory(2*CHR_COUNT* sizeof(struct Vertex));
}

// 自機パラメータ制御 vertices[0-1]利用
void MyShip(){
	static int iMyShipAnime=0; // 8フレームに一度ブリンクさせる変数
	// 上下左右操作はMyShipで処理
	if((pad.Buttons & PSP_CTRL_UP   ) && m_iShip_y>0  ) m_iShip_y-=4;
	if((pad.Buttons & PSP_CTRL_DOWN ) && m_iShip_y<256) m_iShip_y+=4;
	if((pad.Buttons & PSP_CTRL_LEFT ) && m_iShip_x>0  ) m_iShip_x-=4;
	if((pad.Buttons & PSP_CTRL_RIGHT) && m_iShip_x<448) m_iShip_x+=4;
	// テクスチャの割り当てやアニメーションを定義
	vertices[MY_SHIP  ].x = m_iShip_x;
	vertices[MY_SHIP+1].x = m_iShip_x+32;
	vertices[MY_SHIP  ].y = m_iShip_y;
	vertices[MY_SHIP+1].y = m_iShip_y+16;
	iMyShipAnime++;
	if(iMyShipAnime%8==0){
		vertices[MY_SHIP  ].u = 0;	vertices[MY_SHIP  ].v = 0;
		vertices[MY_SHIP+1].u = 32;	vertices[MY_SHIP+1].v = 16;
	}else{
		vertices[MY_SHIP  ].u = 32;	vertices[MY_SHIP  ].v = 0;
		vertices[MY_SHIP+1].u = 64;	vertices[MY_SHIP+1].v = 16;
	}
}
// 自機バレット制御関数コール
void MyBullet(){
	static int iBulletNextWait=0;
	// Xキーを押した場合は射出ルーチンを処理
	if((pad.Buttons & PSP_CTRL_CROSS) && iBulletNextWait==0){
		// 空きバレットを走査し、空きがあれば目の前に設置
		for(i=0;i<=9;i++) if(m_iBullet_flag[i]==0) break;
		if(i!=10){
			m_iBullet_flag[i]=1; // バレットを使用中に
			m_iBullet_x[i] = m_iShip_x + 20;
			m_iBullet_y[i] = m_iShip_y;
			iBulletNextWait = 4; // 4フレームのウェイトを入れる
		}
	}
	if(iBulletNextWait!=0) iBulletNextWait--;
	for(i=0;i<=9;i++){
		if(m_iBullet_flag[i]==1){ // 画面に存在している場合は
			m_iBullet_x[i]+=m_iBullet_spd; // 弾速分を加算
			vertices[MY_BULLET+i*2  ].x = m_iBullet_x[i];
			vertices[MY_BULLET+i*2+1].x = m_iBullet_x[i]+16;
			vertices[MY_BULLET+i*2  ].y = m_iBullet_y[i];
			vertices[MY_BULLET+i*2+1].y = m_iBullet_y[i]+16;
			vertices[MY_BULLET+i*2  ].u = 64;
			vertices[MY_BULLET+i*2+1].u = 80;
			vertices[MY_BULLET+i*2  ].v = 0;
			vertices[MY_BULLET+i*2+1].v = 16;
			// 存在している場合のみボスとの当たり判定を求める
			if(m_iBullet_x[i]>=m_iBoss_x[m_iStage] && m_iBullet_x[i]<=m_iBoss_x[m_iStage]+60 &&
			   m_iBullet_y[i]+12>=m_iBoss_y[m_iStage] && m_iBullet_y[i]<=m_iBoss_y[m_iStage]+60){
				// エフェクト0-5を走査して空きにセット
				for(j=0;j<=4;j++){
					if(m_iEffect_time[j]<0){
						m_iEffect_x[j] = m_iBullet_x[i];
						m_iEffect_y[j] = m_iBullet_y[i];
						m_iEffect_time[j] = 31; // アニメーションフレームセット
						break; // 空きを発見したらループ脱出
					}
				}
				m_iBullet_x[i]=512; // 衝突した場合は画面外までバレットを飛ばす
				m_iBoss_hp[m_iStage]--; // ボスの耐久値をマイナス1
			}
		}
		// 画面外に消えたら空きバレットに降格
		if(m_iBullet_x[i]>512) m_iBullet_flag[i]=0;
	}
}

// ボスの射出する敵弾関連ルーチン
void MyGenerateEnemyBullet(double x,double y,int type,double spd,int num,int hitbox,int power,int autodec){
	// 空きを走査する
	for(i=0;i<MY_ENEMY_BULLET_MAX;i++) if(m_iEnemyBullet_Type[i]==-1) break;
	// 空きスロットに敵弾ステータスをセット
	m_dEnemyBullt_x[i] = x;
	m_dEnemyBullt_y[i] = y;
	m_iEnemyBullet_Type[i] = type;
	m_dENemyBullet_Spd[i] = spd;
	m_iEnemyBullet_Num[i] = num;
	m_iEnemyBullet_Hitbox[i] = hitbox;
	m_iEnemyBullet_Power[i] = power;
	m_iEnemyBullet_Autodec[i] = autodec;
}
void MyEnemyBullet(){
	for(i=0;i<MY_ENEMY_BULLET_MAX;i++){
		if(m_iEnemyBullet_Type[i]==-1) continue; // 無駄弾スキップ
		// 敵弾タイプに応じて軌道を変える
		switch(m_iEnemyBullet_Type[i]){
			default:
			case 0:
				break;
			case 1:
				m_dEnemyBullt_x[i]-=m_dENemyBullet_Spd[i];
				break;
		}
		// 自機の当たり判定を処理
		if((m_iEnemyBullet_Hitbox[i] & 0x01) && m_iShip_x+16>(int)m_dEnemyBullt_x[i] && m_iShip_x+16<(int)m_dEnemyBullt_x[i]+16
		 && m_iShip_y+8>(int)m_dEnemyBullt_y[i] && m_iShip_y+8<(int)m_dEnemyBullt_y[i]+16){
			m_iGameoverFlag = 1; // ゲームオーバーフラグを立てる
			return;
		}
		// 自機バレットの当たり判定を処理
		for(j=0;j<=9;j++){
			if(m_iBullet_flag[j]==0) continue; // 未射出バレットは処理しない
			if((m_iEnemyBullet_Hitbox[i] & 0x02) && m_iBullet_x[j]+8>(int)m_dEnemyBullt_x[i] && m_iBullet_x[j]+8<(int)m_dEnemyBullt_x[i]+16
			 && m_iBullet_y[j]+8>(int)m_dEnemyBullt_y[i] && m_iBullet_y[j]+8<(int)m_dEnemyBullt_y[i]+16){
				m_iEnemyBullet_Power[i]--; // 敵弾の耐久度を-1する
				m_iBullet_x[j]=512; // 画面外まで自機バレットを移動させて消去
			}
		}
		// 自動減衰処理も行う
		m_iEnemyBullet_Power[i] -= m_iEnemyBullet_Autodec[i];
		// 敵弾の耐久性が0未満になったらタイプ-1に戻す
		if(m_iEnemyBullet_Power[i]<0){
			m_iEnemyBullet_Type[i]=-1;
			// エフェクト0-5を走査して空きにセット
			for(j=0;j<=4;j++){
				if(m_iEffect_time[j]<0){
					m_iEffect_x[j] = (int)m_dEnemyBullt_x[i];
					m_iEffect_y[j] = (int)m_dEnemyBullt_y[i];
					m_iEffect_time[j] = 31; // アニメーションフレームセット
					break; // 空きを発見したらループ脱出
				}
			}
		}

		// 弾が画面外に出たらタイプを-1に戻す
		if(m_dEnemyBullt_x[i]<-32 || m_dEnemyBullt_x[i]>512 ||
		m_dEnemyBullt_y[i]<-32 || m_dEnemyBullt_y[i]>320)
			m_iEnemyBullet_Type[i]=-1;
		// スプライト値をセット
		vertices[MY_ENEMY_BULLET+i*2  ].x = (int)m_dEnemyBullt_x[i];
		vertices[MY_ENEMY_BULLET+i*2+1].x = (int)m_dEnemyBullt_x[i]+16;
		vertices[MY_ENEMY_BULLET+i*2  ].y = (int)m_dEnemyBullt_y[i];
		vertices[MY_ENEMY_BULLET+i*2+1].y = (int)m_dEnemyBullt_y[i]+16;
		vertices[MY_ENEMY_BULLET+i*2  ].u = 64 + m_iEnemyBullet_Num[i]*16;
		vertices[MY_ENEMY_BULLET+i*2+1].u = 80 + m_iEnemyBullet_Num[i]*16;
		vertices[MY_ENEMY_BULLET+i*2  ].v = 16;
		vertices[MY_ENEMY_BULLET+i*2+1].v = 32;
	}
}

// ボス制御ルーチン一式
void MyInitBoss(){
	switch(m_iStage){
		default:
		case 0: // ステージ1の初期値
			m_iBoss_MoveY = m_iBoss_spd[m_iStage];
			break;
	}
}
void MyBoss(){
	// ステージ個別にパラメータ参照
	vertices[MY_BOSS  ].x = m_iBoss_x[m_iStage];
	vertices[MY_BOSS+1].x = m_iBoss_x[m_iStage] + 64;
	vertices[MY_BOSS  ].y = m_iBoss_y[m_iStage];
	vertices[MY_BOSS+1].y = m_iBoss_y[m_iStage] + 64;
	vertices[MY_BOSS  ].u = m_iBoss_u[m_iStage];
	vertices[MY_BOSS+1].u = m_iBoss_u[m_iStage] + 64;
	vertices[MY_BOSS  ].v = m_iBoss_v[m_iStage];
	vertices[MY_BOSS+1].v = m_iBoss_v[m_iStage] + 64;
	switch(m_iStage){
		default:
		case 0: // ステージ1
			m_iBoss_y[m_iStage] += m_iBoss_MoveY;
			// 画面端で反転
			if(m_iBoss_y[m_iStage]< 0 || m_iBoss_y[m_iStage]>208) m_iBoss_MoveY = -m_iBoss_MoveY;
			// 30フレーム間隔で交互に電源ボタンと開閉ボタンから敵弾を射出
			if(m_ulFrameCount%60==0)
				MyGenerateEnemyBullet(m_iBoss_x[m_iStage],m_iBoss_y[m_iStage]+4 ,1,4.0f,0,1,1,0);
			if((30+m_ulFrameCount)%60==0)
				MyGenerateEnemyBullet(m_iBoss_x[m_iStage],m_iBoss_y[m_iStage]+44,1,4.0f,0,1,1,0);
			// テストコード
			if(m_ulFrameCount%15==0){
				MyGenerateEnemyBullet(m_iBoss_x[m_iStage]-32,m_iBoss_y[m_iStage]+32,0,0.0f,1,3,120,1);
			}

			break;
	}
}

// エフェクト制御ルーチン
void MyEffect(){
	for(i=0;i<=4;i++){
		if(m_iEffect_time[i]<0){ // フラグ0以下のときは画面外に
			m_iEffect_x[i] = 512;
			vertices[MY_EFFECT+i*2  ].x = m_iEffect_x[i];
			vertices[MY_EFFECT+i*2+1].x = m_iEffect_x[i]+16;
		}else{ // アニメーションフレームが-1以外の場合は描画
			vertices[MY_EFFECT+i*2  ].x = m_iEffect_x[i];
			vertices[MY_EFFECT+i*2+1].x = m_iEffect_x[i]+16;
			vertices[MY_EFFECT+i*2  ].y = m_iEffect_y[i];
			vertices[MY_EFFECT+i*2+1].y = m_iEffect_y[i]+16;
			vertices[MY_EFFECT+i*2  ].u = 48 - ((m_iEffect_time[i] / 8) *16);
			vertices[MY_EFFECT+i*2+1].u = 64 - ((m_iEffect_time[i] / 8) *16);
			vertices[MY_EFFECT+i*2  ].v = 16;
			vertices[MY_EFFECT+i*2+1].v = 32;
			m_iEffect_time[i]-=2; // フレームを2ずつ減らす
			m_iEffect_x[i]+=2; // 衝突位置から爆風が反動で前進
		}
	}
}

// メインルーチン
int main() {
	MyInitGu();		// gu初期化コール
	MyInitParam();	// 変数初期化コール
	MyInitBoss();	// ボス初期化コール(暫定)。最終的にはステージセレクト後に呼び出す予定。
	while(1){
		sceCtrlReadBufferPositive(&pad, 1); // パッド情報セット
		sceGuStart(GU_DIRECT,list);
		sceGuClear(GU_COLOR_BUFFER_BIT); // 毎回描画クリア
		MyShip();	// 自機制御関数コール
		MyBullet();	// 自機バレット制御関数コール
		MyBoss();	// ボス制御関数コール
		MyEffect(); // エフェクト制御関数コール
		MyEnemyBullet(); // 敵弾処理関数コール
		m_ulFrameCount++; // フレームカウント変数インクリメント
		// ループ末尾で描画処理を実行
		sceGuDrawArray(GU_SPRITES,GU_TEXTURE_16BIT|GU_COLOR_4444|GU_VERTEX_16BIT|GU_TRANSFORM_2D, CHR_COUNT*2, 0, vertices);
		sceGuFinish(); sceGuSync(0,0); sceDisplayWaitVblankStart(); sceGuSwapBuffers();
	}
	return 0;
}
